/*
 * Copyright (C) 2008  Anthony Liguori <anthony@codemonkey.ws>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <gdk/gdkkeysyms.h>
#include "x_keymap.h"
#include "utils.h"
#include "vnc_keycodes.h"

/*
 * This table is taken from QEMU x_keymap.c, under the terms:
 *
 * Copyright (c) 2003 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* Mapping is offset by 97 to save space */
static const uint8_t x_keycode_to_pc_keycode_table[61] = {
	0xc7,      /*  97  Home   */
	0xc8,      /*  98  Up     */
	0xc9,      /*  99  PgUp   */
	0xcb,      /* 100  Left   */
	0x4c,        /* 101  KP-5   */
	0xcd,      /* 102  Right  */
	0xcf,      /* 103  End    */
	0xd0,      /* 104  Down   */
	0xd1,      /* 105  PgDn   */
	0xd2,      /* 106  Ins    */
	0xd3,      /* 107  Del    */
	0x9c,      /* 108  Enter  */
	0x9d,      /* 109  Ctrl-R */
	0x0,       /* 110  Pause  */
	0xb7,      /* 111  Print  */
	0xb5,      /* 112  Divide */
	0xb8,      /* 113  Alt-R  */
	0xc6,      /* 114  Break  */
	0x0,         /* 115 */
	0x0,         /* 116 */
	0x0,         /* 117 */
	0x0,         /* 118 */
	0x0,         /* 119 */
	0x0,         /* 120 */
	0x0,         /* 121 */
	0x0,         /* 122 */
	0x0,         /* 123 */
	0x0,         /* 124 */
	0x0,         /* 125 */
	0x0,         /* 126 */
	0x0,         /* 127 */
	0x0,         /* 128 */
	0x79,         /* 129 Henkan */
	0x0,         /* 130 */
	0x7b,         /* 131 Muhenkan */
	0x0,         /* 132 */
	0x7d,         /* 133 Yen */
	0x0,         /* 134 */
	0x0,         /* 135 */
	0x47,         /* 136 KP_7 */
	0x48,         /* 137 KP_8 */
	0x49,         /* 138 KP_9 */
	0x4b,         /* 139 KP_4 */
	0x4c,         /* 140 KP_5 */
	0x4d,         /* 141 KP_6 */
	0x4f,         /* 142 KP_1 */
	0x50,         /* 143 KP_2 */
	0x51,         /* 144 KP_3 */
	0x52,         /* 145 KP_0 */
	0x53,         /* 146 KP_. */
	0x47,         /* 147 KP_HOME */
	0x48,         /* 148 KP_UP */
	0x49,         /* 149 KP_PgUp */
	0x4b,         /* 150 KP_Left */
	0x4c,         /* 151 KP_ */
	0x4d,         /* 152 KP_Right */
	0x4f,         /* 153 KP_End */
	0x50,         /* 154 KP_Down */
	0x51,         /* 155 KP_PgDn */
	0x52,         /* 156 KP_Ins */
	0x53,         /* 157 KP_Del */
};

/* This table was put together by VirtualBox.  It's based on the information in
 * /usr/share/X11/xkb/keycodes/evdev using the x_keycode_to_pc_keycode table
 * to get the pc keycodes.
 */

static const uint8_t evdev_keycode_to_pc_keycode[61] = {
	0,    /* 97 unknown */
	0,    /* 98 unknown */
	0,    /* 99 unknown */
	0,    /* 100 unknown */
	0,    /* 101 unknown */
	0,    /* 102 unknown */
	0,    /* 103 unknown */
	0x9c, /* 104 KPEN */
	0x9d, /* 105 RCTL */
	0xb5, /* 106 KPDV */
	0xb7, /* 107 PRSC */
	0xb8, /* 108 RALT */
	0,    /* 109 unknown */
	0xc7, /* 110 HOME */
	0xc8, /* 111 UP */
	0xc9, /* 112 PGUP */
	0xcb, /* 113 LEFT */
	0xcd, /* 114 RGHT */
	0xcf, /* 115 END */
	0xd0, /* 116 DOWN */
	0xd1, /* 117 PGDN */
	0xd2, /* 118 INS */
	0xd3, /* 119 DELE */
	0,    /* 120 unknown */
	0,    /* 121 unknown */
	0,    /* 122 unknown */
	0,    /* 123 unknown */
	0,    /* 124 unknown */
	0,    /* 125 unknown */
	0,    /* 126 unknown */
	0,    /* 127 unknown */
	0,    /* 128 unknown */
	0,    /* 129 unknown */
	0,    /* 130 unknown */
	0,    /* 131 unknown */
	0x7d, /* 132 AE13 */
	0,    /* 133 unknown */
	0,    /* 134 unknown */
	0,    /* 135 unknown */
	0,    /* 136 unknown */
	0,    /* 137 unknown */
	0,    /* 138 unknown */
	0,    /* 139 unknown */
	0,    /* 140 unknown */
	0,    /* 141 unknown */
	0,    /* 142 unknown */
	0,    /* 143 unknown */
	0,    /* 144 unknown */
	0,    /* 145 unknown */
	0,    /* 146 unknown */
	0,    /* 147 unknown */
	0,    /* 148 unknown */
	0,    /* 149 unknown */
	0,    /* 150 unknown */
	0,    /* 151 unknown */
	0,    /* 152 unknown */
	0,    /* 153 unknown */
	0,    /* 154 unknown */
	0,    /* 155 unknown */
	0,    /* 156 unknown */
	0,    /* 157 unknown */
};

/* keycode translation for sending ISO_Left_Send
 * to vncserver
 */
static struct {
	GdkKeymapKey *keys;
	gint n_keys;
	guint keyval;
} untranslated_keys[] = {{NULL, 0, GDK_Tab}};

static unsigned int ref_count_for_untranslated_keys = 0;

/* As best as I can tell, windows passes PC scan codes.  This code definitely
 * won't compile on windows so let's put in a guard anyway */

#ifndef WIN32
#include <gdk/gdkx.h>
#include <X11/XKBlib.h>
#include <stdbool.h>
#include <string.h>

#define STRPREFIX(a,b) (strncmp((a),(b),strlen((b))) == 0)

static gboolean check_for_evdev(void)
{
	XkbDescPtr desc;
	gboolean has_evdev = FALSE;
	const gchar *keycodes;

	desc = XkbGetKeyboard(GDK_DISPLAY(), XkbGBN_AllComponentsMask,
			      XkbUseCoreKbd);
	if (desc == NULL || desc->names == NULL)
		return FALSE;

	keycodes = gdk_x11_get_xatom_name(desc->names->keycodes);
	if (keycodes == NULL)
		g_warning("could not lookup keycode name\n");
	else if (STRPREFIX(keycodes, "evdev_"))
		has_evdev = TRUE;
	else if (!STRPREFIX(keycodes, "xfree86_"))
		g_warning("unknown keycodes `%s', please report to gtk-vnc-devel\n",
			  keycodes);

	XkbFreeClientMap(desc, XkbGBN_AllComponentsMask, True);

	return has_evdev;
}
#else
static gboolean check_for_evdev(void)
{
	return FALSE;
}
#endif

const uint8_t const *x_keycode_to_pc_keycode_map(void)
{
	if (check_for_evdev()) {
		GVNC_DEBUG("Using evdev keycode mapping");
		return evdev_keycode_to_pc_keycode;
	} else {
		GVNC_DEBUG("Using xfree86 keycode mapping");
		return x_keycode_to_pc_keycode_table;
	}
}

uint16_t x_keycode_to_pc_keycode(const uint8_t const *keycode_map,
				 uint16_t keycode)
{
	if (keycode == GDK_Pause)
		return VKC_PAUSE;

	if (keycode < 9)
		return 0;

	if (keycode < 97)
		return keycode - 8; /* just an offset */

	if (keycode < 158)
		return keycode_map[keycode - 97];

	if (keycode == 208) /* Hiragana_Katakana */
		return 0x70;

	if (keycode == 211) /* backslash */
		return 0x73;

	return 0;
}

/* Set the keymap entries */
void x_keymap_set_keymap_entries()
{
	size_t i;
	if (ref_count_for_untranslated_keys == 0)
		for (i = 0; i < sizeof(untranslated_keys) / sizeof(untranslated_keys[0]); i++)
			gdk_keymap_get_entries_for_keyval(gdk_keymap_get_default(),
							  untranslated_keys[i].keyval,
							  &untranslated_keys[i].keys,
							  &untranslated_keys[i].n_keys);
	ref_count_for_untranslated_keys++;
}

/* Free the keymap entries */
void x_keymap_free_keymap_entries()
{
	size_t i;

	if (ref_count_for_untranslated_keys == 0)
		return;

	ref_count_for_untranslated_keys--;
	if (ref_count_for_untranslated_keys == 0)
		for (i = 0; i < sizeof(untranslated_keys) / sizeof(untranslated_keys[0]); i++)
			g_free(untranslated_keys[i].keys);

}

/* Get the keyval from the keycode without the level. */
guint x_keymap_get_keyval_from_keycode(guint keycode, guint keyval)
{
	size_t i;
	for (i = 0; i < sizeof(untranslated_keys) / sizeof(untranslated_keys[0]); i++) {
		if (keycode == untranslated_keys[i].keys[0].keycode) {
			return untranslated_keys[i].keyval;
		}
	}

	return keyval;
}
/*
 * Local variables:
 *  c-indent-level: 8
 *  c-basic-offset: 8
 *  tab-width: 8
 * End:
 */
